;;; $Id: atomic-tree.scm,v 1.5 2006/01/05 22:24:09 leavens Exp $
;;; Trees of atomic items, using the grammar:
;;;    <(tree-of  T)> ::= ( {<(atomic-expression-of T)>}* )
;;;    <(atomic-expression-of T)> ::= <T> | <(tree-of  T)>
;;; where the T's are required to be atomic and not null.
;;; This is equivalent to:
;;;    <(tree-of T)> ::= () 
;;;                |  ( <T> . <(tree-of T)> )
;;;                |  ( <(tree-of T)> . <(tree-of T)> )
;;; See the file atomic-exp.scm for a grammar that doesn't use a tree ADT.

(module atomic-tree (lib "typedscm.ss" "typedscm")

(provide tree-of tree tree-empty tree-add tree-add-atom tree-add-tree
         tree-null? tree-first-atomic? tree-first-tree?
         tree-first tree-first-atom tree-first-tree tree-rest parse-tree
         atomic-expression-of atom->atomic-expression tree->atomic-expression
         atomic-expression-atom? atomic-expression-tree?
         atomic-expression->atom atomic-expression->tree
         parse-atomic-expression atomic?)

(deftype tree-of
  (forall (T) (-> ((type-predicate-for T)) (type-predicate-for (tree-of T)))))
(deftype tree
  (forall (T) (-> ((atomic-expression-of T)...) (tree-of T))))
(deftype tree-empty
  (forall (T) (-> () (tree-of T))))
(deftype tree-add
  (forall (T) (-> ((atomic-expression-of T) (tree-of T)) (tree-of T))))
(deftype tree-add-atom
  (forall (T) (-> (T (tree-of T)) (tree-of T))))
(deftype tree-add-tree
  (forall (T) (-> ((tree-of T) (tree-of T)) (tree-of T))))
(deftype tree-null?
  (forall (T) (-> ((tree-of T)) boolean)))
(deftype tree-first-atomic?
  (forall (T) (-> ((tree-of T)) boolean)))
(deftype tree-first-tree?
  (forall (T) (-> ((tree-of T)) boolean)))
(deftype tree-first
  (forall (T) (-> ((tree-of T)) (atomic-expression-of T))))
(deftype tree-first-atom
  (forall (T) (-> ((tree-of T)) T)))
(deftype tree-first-tree
  (forall (T) (-> ((tree-of T)) (tree-of T))))
(deftype tree-rest
  (forall (T) (-> ((tree-of T)) (tree-of T))))
(deftype parse-tree (-> (datum) (tree-of datum)))

(deftype atomic-expression-of
  (forall (T) (-> ((type-predicate-for T))
                  (type-predicate-for (atomic-expression-of T)))))
(deftype atom->atomic-expression
  (forall (T) (-> (T) (atomic-expression-of T))))
(deftype tree->atomic-expression
  (forall (T) (-> ((tree-of  T)) (atomic-expression-of T))))
(deftype atomic-expression-atom?
  (forall (T) (-> ((atomic-expression-of T)) boolean)))
(deftype atomic-expression-tree?
  (forall (T) (-> ((atomic-expression-of T)) boolean)))
(deftype atomic-expression->atom
  (forall (T) (-> ((atomic-expression-of T)) T)))
(deftype atomic-expression->tree
  (forall (T) (-> ((atomic-expression-of T)) (tree-of T))))
(deftype parse-atomic-expression
  (forall (T S)
    (all-of (-> ((list-of T)) (atomic-expression-of T))
            (-> (S) (atomic-expression-of S))
            (-> (datum) (atomic-expression-of datum)))))

(require (lib "atomic-mod.scm" "lib342"))

(defrep
  (forall (T) (tree-of T) (list-of (atomic-expression-of T)))
  (forall (T) (atomic-expression-of T) datum))  ; run-time checked

(define tree-of
  (forall (T?)
    (list-of (atomic-expression-of T?))))
                  
(define tree
  (lambda aes
    aes))

(define tree-empty
  (lambda ()
    '()))

(define tree-add
  cons)

(define tree-add-atom
  (lambda (e tr)
    (tree-add (atom->atomic-expression e) tr)))

(define tree-add-tree
  (lambda (e tr)
    (tree-add (tree->atomic-expression e) tr)))

(define tree-null?
  null?)

(define tree-first-atomic?
  (lambda (tr)
    (atomic-expression-atom? (car tr))))

(define tree-first-tree?
  (lambda (tr)
    (atomic-expression-tree? (car tr))))

(define tree-first
  car)

(define tree-first-atom
  (lambda (tr)
    (atomic-expression->atom (tree-first tr))))

(define tree-first-tree
  (lambda (tr)
    (atomic-expression->tree (tree-first tr))))

(define tree-rest
  cdr)

(define parse-tree
  (lambda (d)
    (test-type
      (list-of datum?)
      (if (list? d)
          (map parse-atomic-expression d)
          (error "parse-tree: bad syntax: " d)))))
  
(define atomic-expression-of
  (forall (T?)
    (has-type-trusted
      (forall (T) (type-predicate-for T))
      (lambda (d)
        (or (atomic? d)
            ((list-of (atomic-expression-of T?)) d))))))
 
(define atom->atomic-expression
  (lambda (t)
    (if (and (atomic? t) (not (null? t)))
        (has-type datum t)
        (error
         "atom->atomic-expression: argument must be atomic and not null, but is: "
         t))))

(define tree->atomic-expression
  (lambda (tr)
    (if (and (list? tr) (atomic-expression-tree? tr))
        (has-type datum tr)
        (error "list->atomic-expression: improper tree argument: " tr))))

(define atomic-expression-atom?
    (lambda (exp)
      (and (atomic? exp) (not (null? exp)))))
      
(define atomic-expression-tree?
  (lambda (exp)
    (or (null? exp) (pair? exp))))

(define atomic-expression->atom
  (lambda (exp)
    (if (atomic-expression-atom? exp)
        exp
        (error "atomic-expression->atom: atomic-expression is a tree " exp))))

(define atomic-expression->tree
  (lambda (exp)
    (if (atomic-expression-tree? exp)
        (test-type (tree-of datum?) exp)
        (error "atomic-expression->atom: atomic-expression is not a tree "
               exp))))

(define parse-atomic-expression
  (has-type-trusted
    (forall (s t)
      (all-of (-> ((list-of s)) datum) (-> (t) datum) (-> (datum) datum)))
    (lambda (exp)
      (cond
       ((list? exp) (map parse-atomic-expression exp))
       ((atomic? exp) (atom->atomic-expression exp))
       (else (error "parse-atomic-expression: bad syntax: " exp))))))

) ;; end module
