;;; $Id: tc-env.scm,v 1.8 2006/02/05 20:01:15 leavens Exp $
;;; Copyright (C) 2006 Iowa State University
;;;
;;; This file is part of Typedscm.
;;;
;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Lesser General Public License
;;; as published by the Free Software Foundation; either version 2.1,
;;; of the License, or (at your option) any later version.
;;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Lesser General Public License for more details.
;;;
;;; You should have received a copy of the GNU Lesser General Public
;;; License along with Typedscm; see the file LesserGPL.txt.  If not,
;;; write to the Free Software Foundation, Inc., 51 Franklin St, Fifth
;;; Floor, Boston, MA 02110-1301 USA.

;;; Environments for the Scheme type checker.
;;; AUTHORS: Curtis Clifton, Gary T. Leavens, and Brian Dorn

;;;----------------------------------------------------------------------------
;;;
;;; An environment is an object that associates variables to values.

(module tc-env (lib "typedscm.ss" "typedscm")

(provide tc:env-of tc:env-empty
	tc:env-update tc:env-update* tc:env-join tc:env-join*
	tc:env-bound? tc:env-value
	tc:env-map tc:env-filter tc:env-for-each
	tc:assoc-list->env
	tc:environment?	tc:env-names tc:env-vals)

(deftype tc:env-of
   (forall (K V)
     (-> ((type-predicate-for K) (type-predicate-for V))
	 (type-predicate-for (tc:env-of K V)))))
(deftype tc:env-empty
  (forall (K V) (-> (number ...) (tc:env-of K V))))
(deftype tc:env-update
  (forall (K V) (-> (K V (tc:env-of K V)) (tc:env-of K V))))
(deftype tc:env-update*
  (forall (K V)
     (-> ((list-of K) (list-of V) (tc:env-of K V)) (tc:env-of K V))))
(deftype tc:env-join
  (forall (K V) (-> ((tc:env-of K V) (tc:env-of K V)) (tc:env-of K V))))
(deftype tc:env-bound?
  (forall (K V) (-> (K (tc:env-of K V)) boolean)))
(deftype tc:env-value
  (forall (K V) (-> (K (tc:env-of K V)) V)))
(deftype tc:env-map
  (forall (K V S T) (-> ((-> (K S) T) (tc:env-of K S)) (tc:env-of K T))))
(deftype tc:env-filter
  (forall (K V) (-> ((-> (K V) boolean) (tc:env-of K V)) (tc:env-of K V))))
(deftype tc:env-for-each
  (forall (K V) (-> ((-> (K V) void) (tc:env-of K V)) void)))
(deftype tc:assoc-list->env
  (forall (K V) (-> ((list-of (pair-of K V))) (tc:env-of K V))))
(deftype tc:environment? (type-predicate-for (tc:env-of symbol datum)))
(deftype tc:env-names
  (forall (S T) (-> ((tc:env-of S T)) (list-of S))))
(deftype tc:env-vals
  (forall (S T) (-> ((tc:env-of S T)) (list-of T))))

(require (lib "tc-hash-table.scm" "typedscm"))

;;; Environments are represented by eq hash tables
(defrep (forall (K V) (tc:env-of K V) (tc:hash-table K V)))

(define tc:env-of
  (has-type-trusted
   (forall (K V)
     (-> ((type-predicate-for K) (type-predicate-for V))
	 (type-predicate-for (tc:hash-table K V))))
   (forall (K? V?)
     (lambda (d)
	(tc:hash-table? d)))))

(define tc:env-empty
  (let ((empty-hash-tab (tc:make-hash-table)))
    (lambda ()
      ;; ENSURES: result is an empty environment
      empty-hash-tab)))

(define tc:env-update 
  (lambda (var val env)
    ;; ENSURES: result is a copy of env but with var associated to val
    (tc:hash-table-associate var val env)))

(define tc:env-update*
  (lambda (vars vals env)
    ;; ENSURES: result is a copy of env but with each element of vars
    ;;          associated to the corresponding element of vals
    (tc:hash-table-associate* vars vals env)))

(define tc:env-join  
  (lambda (env1 env2)
    ;; ENSURES: result is a new environment containing the bindings
    ;; in env1 and env2 (env1 has precedence)
    (tc:hash-table-join env1 env2)))

(define tc:env-join*
  (lambda (envs)
    ;; ENSURES: result is a new environment containing the bindings
    ;; in all environments in envs where the precedence is given
    ;; in order of occurence in the list
    (if (null? envs)
	(tc:env-empty)
	(tc:env-join*-helper (car envs) (cdr envs)))))

(define tc:env-join*-helper
  (lambda (env1 envs)
    (if (null? envs)
	env1
	(tc:env-join*-helper
	 (tc:env-join env1 (car envs))
	 (cdr envs)))))		     

(define tc:env-bound? 
  (lambda (var env)
    ;; ENSURES: result is true iff var is bound in env
    (tc:hash-table-defined? var env)))

(define tc:env-value
  (let ((error-cont
	 (lambda () (error "tc:env-value called with key not in table!"))))
    (lambda (var env)
      ;; REQUIRES: var has an association in env
      ;; ENSURES: result is the value associated to var in env
      (tc:hash-table-value var env error-cont))))

(define tc:env-map
  (lambda (p env)
    ;; EFFECT: returns an environment that is like env but with
    ;;         every (k, v) association in env replaced with (k, (p k v)).
    (tc:hash-table-map p env)))

(define tc:env-filter
  (lambda (pred env)
    ;; ENSURES: result is an environment that is like env, but which
    ;;         only contains the bindings for which pred is true.
    (tc:hash-table-filter pred env)))

(define tc:env-for-each
  (lambda (p env)
    ;; EFFECT: calls p on every pair k and v
    ;;         for which k is associated to v in env.
    (tc:hash-table-for-each p env)))

;;; The following assists in converting the global var types to environments
(define tc:assoc-list->env
   (lambda (alist)
     (has-type-trusted
       (forall (K V) (tc:hash-table K V))
       (tc:env-update* (map car alist) (map cdr alist)
		       (tc:env-empty)))))

;;; The following are used in the type checker
(define tc:environment?
  (lambda (d)
    ((tc:env-of symbol? datum?) d)))

(define tc:env-names
  (lambda (env)
    (let ((result '()))
      (begin
	(tc:env-for-each
	 (lambda (key val)
	   (set! result (cons key result)))
	 env)
	result))))

(define tc:env-vals
  (lambda (env)
    (let ((result '()))
      (begin
	(tc:env-for-each
	 (lambda (key val)
	   (set! result (cons val result)))
	 env)
	result))))

) ;; end module
